local json = require("json")
local const = require("const")
local utils = require("utils")
local driver_utils = require("driver_utils")
local pl_path = require("pl.path")
local request_method = ngx.var.request_method

utils.check_login()
utils.check_permission("DriverManage")

local function generate_random_num()
    return math.random(1,999999999)
end

--0: error, 1: ipv4, 2: ipv6, 3: just string
local function get_ip_type(ip)
    -- must pass in a string value
    if ip == nil or type(ip) ~= "string" then
        return 0
    end

    -- check for format 1.11.111.111 for ipv4
    local chunks = {ip:match("(%d+)%.(%d+)%.(%d+)%.(%d+)")}
    if (#chunks == 4) then
        for _,v in pairs(chunks) do
            if (tonumber(v) < 0 or tonumber(v) > 255) then
                return 0
            end
        end
        return 1
    else
        return 0
    end

    -- check for ipv6 format, should be 8 'chunks' of numbers/letters
    local _, chunks = ip:gsub("[%a%d]+%:?", "")
    if chunks == 8 then
        return 2
    end

    -- if we get here, assume we've been given a random string
    return 3
end

local function is_serverId_valid(cfg, id)
    for k, v in pairs(cfg["serverList"]) do
        if id == v["serverId"] then
            return true
        end
    end
    return false
end

local function is_serverId_repeated(cfg, cur_index, cur_val)
    for k2, v2 in pairs(cfg["serverList"]) do
        if k2 > cur_index and cur_val == v2["serverId"] then
            return true
        end
    end
    return false
end

local function check_modbus_tcp(name, ip, port)

    if name ~= nil and type(name) ~= "string" then
        return false, "name must be string type"
    end

    if ip == nil or get_ip_type(ip) ~= 1 then
        return false, "ip is invalid"
    end

    if port == nil or type(port) ~= "number" or tonumber(port) <= 0 then
        return false, "port is invalid: "
    end

    return true
end

local function check_bitrate(bt)
    local valid_bt = {300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 43000, 56000, 57600, 115200}

    for k , v in ipairs(valid_bt) do
        if v == bt then
            return true
        end
    end

    return false
end

local function check_modbus_rtu(name, serialPort, baudRate, byteSize, parity, stopBits)

    if name ~= nil and type(name) ~= "string" then
        return false, "name must be string type"
    end

    if serialPort == nil or type(serialPort) ~= "string" then
        return false, "serialPort must be string type"
    end

    if baudRate == nil or type(baudRate) ~= "number" then
        return false, "serialPort must be string type"
    end

    if false == check_bitrate(baudRate) then
        return false, "baudRate is invalid"
    end

    if not (byteSize == 7 or byteSize == 8) then
        return false, "byteSize is invalid"
    end

    if not (parity == 1 or parity == 2 or parity == 3) then
        return false, "parity is invalid"
    end

    if not (stopBits == 1 or stopBits == 1.5 or stopBits == 2) then
        return false, "stopBits is invalid"
    end
end

local function is_ip_port_repeated(name, ip, port, cfg, cur_index)
    for k, v in pairs(cfg["serverList"]) do
        if (k > cur_index) and ((ip == v["ip"] and port == v["port"])
            or (name ~= nil and name == v["name"])) then
            return true
        end
    end
    return false
end

local function is_serial_port_repeated(name, serial, cfg, cur_index)
    for k, v in pairs(cfg["serverList"]) do
        if (k > cur_index and v["protocol"] == "RTU") and
            ((serial == v["serialPort"]..v["baudRate"]..v["byteSize"]..v["parity"]..v["stopBits"])
                or (name ~= nil and name == v["name"])) then
            return true
        end
    end
    return false
end

local function check_cfg_modbus_server_list(cfg)
    for k , v in pairs(cfg["serverList"]) do
        local server_id = v["serverId"]
        if server_id == nil or type(server_id) ~= "string" or string.len(server_id) <= 0 then
            cfg["serverList"][k]["serverId"] = "modbus_server_id_"..generate_random_num()
        elseif is_serverId_repeated(cfg, k, server_id) then
            return false, "repeated server id"
        end

        local protocol = v["protocol"]
        if protocol == "TCP" then
            local code, msg = check_modbus_tcp(v["name"], v["ip"], v["port"])
            if false == code then
                return code, msg
            end

            if is_ip_port_repeated(v["name"], v["ip"], v["port"], cfg, k) then
                return false, "ip&port or server name is repeated "
            end

        elseif protocol == "RTU" then
            local code, msg = check_modbus_rtu(v["name"], v["serialPort"], v["baudRate"],
                                            v["byteSize"], v["parity"], v["stopBits"])
            if false == code then
                return code, msg
            end

            if is_serial_port_repeated(v["name"],
                v["serialPort"]..v["baudRate"]..v["byteSize"]..v["parity"]..v["stopBits"], cfg, k) then
                return false, "client name or serial port is repeated"
            end
        else
            return false, "protocol is invalid"
        end
    end
    return true, "", cfg
end

local function is_pk_dn_repeated(deviceList, pk, dn, cur_index)

    for k, v in pairs(deviceList) do
        if k > cur_index and v["productKey"] == pk and v["deviceName"] == dn then
            return true, "pk+dn is repeated."
        end
    end

    return false, ""
end

local function is_slaveId_repeated(cfg)
    local tmp_tab = {}
    for k1, server in pairs(cfg["serverList"]) do
        for k2, client in pairs(cfg["deviceList"]) do
            if client["deviceConfig"]["serverId"] == server["serverId"] then
                if server["protocol"] == "TCP" then
                    table.insert(tmp_tab, server["ip"]..server["port"]..client["deviceConfig"]["slaveId"])
                else
                    table.insert(tmp_tab, server["serialPort"]..client["deviceConfig"]["slaveId"])
                end
            end
        end
    end

    for k3, v3 in ipairs(tmp_tab) do
        for k4, v4 in ipairs(tmp_tab) do
            if k4 > k3 and v4 == v3 then
                return false, "slave id is repeated: "..v4
            end
        end
    end

    return true
end

local function check_cfg_modbus_device_list(cfg)

    for k , v in pairs(cfg["deviceList"]) do
        productKey = v["productKey"]
        deviceName = v["deviceName"]
        deviceConfig = v["deviceConfig"]
        serverId = v["deviceConfig"]["serverId"]
        slaveId = v["deviceConfig"]["slaveId"]

        utils.check_pk(productKey)
        utils.check_dn(deviceName)

        local code, msg = is_pk_dn_repeated(cfg["deviceList"], productKey, deviceName, k)
        if code == true then
            return false, msg
        end
        if deviceConfig == nil or type(deviceConfig) ~= "table" then
            return false, "device config is invalid"
        end
        if is_serverId_valid(cfg, serverId) == false then
            return false, "server id is not exist"
        end
        if slaveId == nil or type(slaveId) ~= "number" or slaveId < 1 or slaveId > 247 then
            return false, "slave id is invalid"
        end
    end

    local code, msg = is_slaveId_repeated(cfg)
    if code == false then
        return false, msg
    end

    return true, cfg
end

local function PKCS5Padding(data)
    local ret = nil

    return nil
end

local function PKCS5Padding(data, size)
    local ret = nil
    if data == nil then
        return nil
    end

    local len = #data
    local padding = size - len%size
    local i = padding
    if padding >= 10 then
        padding = padding + 6
    end

    while(i > 0)
    do
       data = data..string.char(padding)
        i = i - 1
    end
    return data
end

local function aesEncrypt(data, secret)
    local ret = nil
    local aes = require "resty.aes"
    local str = require "resty.string"

    local aes_128_cbc_with_iv = assert(aes:new(string.format("%s", secret),
        nil, aes.cipher(256,"cbc"), {iv=string.sub(secret, 1, 16)}))
    local encrypted = aes_128_cbc_with_iv:encrypt(data)
    return ngx.encode_base64(encrypted)
end

local function check_opcua_url_repeated(cfg, cul_index, cur_url)

	for k, v in pairs(cfg["serverList"]) do
		if k > cul_index and v["url"] == cur_url then
			return false, "url is repeated."
		end
	end

	return true
end

local function check_cfg_opcua_server_list(cfg)

    local ok3, msg_ds = utils.get_data(const.data_key.gwconfig.devicesecret)
    for k, v in pairs(cfg["serverList"]) do
        local server_id = v["serverId"]
        if server_id == nil or type(server_id) ~= "string" or string.len(server_id) <= 0 then
            cfg["serverList"][k]["serverId"] = "opcua_server_id_"..generate_random_num()
        elseif is_serverId_repeated(cfg, k, v["serverId"]) then
            return false, "repeated server id"
        end

        local url = v["url"]
        if url == nil or string.sub(url, 1, string.len("opc.tcp://")) ~= "opc.tcp://" then
            return false, "url is invalid"
        end

        if check_opcua_url_repeated(cfg, k, url) == false then
            return false, "opcua url is repeated"
        end

        local userName = v["userName"]
        local password = v["password"]
        if (userName ~= nil and password == nil) or
            (userName == nil and password ~= nil) then
            return false, "user name and password is not in pair"
        end

        if type(userName) ~= "string" or type(password) ~= "string" then
            return false, "user name and password must be string type"
        end

        if v["flag_encrypt_password"] then
            v["password"] = aesEncrypt(password, msg_ds)
            v["flag_encrypt_password"] = nil
        end
        local secPolicy = v["secPolicy"]
        if not (secPolicy == "None" or  secPolicy == "Basic128Rsa15" or  secPolicy == "Basic256") then
            return false, "secure policy is invalid"
        end

        local secMode = v["secMode"]
        if not (secMode == "None" or secMode == "Sign" or secMode == "SignAndEncrypt") then
            return false, "secure mode is invalid"
        end

        local certificate = v["certificate"]
        local privateKey = v["privateKey"]
        if secMode == "SignAndEncrypt" then
            if certificate == nil or pl_path.isfile(certificate) == false then
                return false, "certificate path is not exist."
            end

            if privateKey == nil or pl_path.isfile(privateKey) == false then
                return false, "privateKey path is not exist."
            end
        end
        local timeout = v["timeout"]
        if type(timeout) ~= "number" or timeout < 0 then
            return false, "timeout must > 0"
        end
    end

    return true, "ok", cfg
end

local function check_opcua_object_path_repeated(cfg)
    local tmp_tab = {}
    local tmp = nil
    for k1, server in pairs(cfg["serverList"]) do
        for k2, client in pairs(cfg["deviceList"]) do
            if client["deviceConfig"]["serverId"] == server["serverId"] then
	 	tmp = server["url"]

	    	table.insert(tmp_tab, string.gsub(tmp, "127.0.0.1", "0.0.0.0", 1)..client["deviceConfig"]["displayNamePath"])
            end
        end
    end

    for k3, v3 in ipairs(tmp_tab) do
        for k4, v4 in ipairs(tmp_tab) do
            if k4 > k3 and v4 == v3 then
                return false, "opcua object path is repeated: "..v4
            end
        end
    end

    return true
end

local function check_cfg_opcua_device_list(cfg)

    for k, v in pairs(cfg["deviceList"]) do
        local productKey = v["productKey"]
        utils.check_pk(productKey)

        local deviceName = v["deviceName"]
        utils.check_dn(deviceName)

        local code, msg = is_pk_dn_repeated(cfg["deviceList"], productKey, deviceName, k)
        if code == true then
            return false, msg
        end

        local deviceConfig = v["deviceConfig"]
        if deviceConfig == nil then
            return false, "device config should not be null"
        end

        local serverId = v["deviceConfig"]["serverId"]
        if is_serverId_valid(cfg, serverId) == false then
            return false, "service id is not exist"
        end

        local displayNamePath = v["deviceConfig"]["displayNamePath"]
        if string.find(displayNamePath, "/") == nil then
            return false, "displayNamePath is invalid"
        end
    end

	local ret, msg = check_opcua_object_path_repeated(cfg)
	if ret == false then
		return false, msg
	end

    return true, "ok", cfg
end


local function check_cfg_websocket_server_list(driver_cfg)

    local count = #driver_cfg["serverList"]
    if count > 1 then
        return false, "the count of server list must be 1"
    end

    for k, cfg in pairs(driver_cfg["serverList"]) do
        local ip = cfg["ip"]
        if ip ~= nil and get_ip_type(ip) ~= 1 then
            return false, "ip is invalid"
        end

        local port = cfg["port"]
        if port == nil or type(port) ~= "number" then
            return false, "port must be integer"
        end

        local auth_mode = cfg["auth_mode"]
        if auth_mode == nil or not (auth_mode == 0 or auth_mode == 1 or auth_mode == 2) then
            return false, "auth mode must be 0/1/2"
        end

        local certification = cfg["certification"]
        if auth_mode ~= 0 then
            if certification == nil or pl_path.isfile(certification) == false then
                return false, "when use ssl is ture, you should set the certificate path"
            end
        end
        local privateKey = cfg["privateKey"]
        if auth_mode ~= 0 then
            if privateKey == nil or pl_path.isfile(privateKey) == false then
                return false, "when use ssl is ture, you should set the privateKey path"
            end
        end

        local publicKey = cfg["publicKey"]
        if auth_mode ~= 0 then
            if publicKey == nil or pl_path.isfile(publicKey) == false then
                return false, "when use ssl is ture, you should set the publicKey path"
            end
        end

        local timeout = cfg["timeout"]
        if timeout == nil or type(timeout) ~= "number" or timeout <= 0 then
            return false, "timeout must > 0"
        end

        local is_self_signature_allowed = cfg["is_self_signature_allowed"]
        if auth_mode == 2 then
            if is_self_signature_allowed == nil or type(is_self_signature_allowed) ~= "boolean" then
                return false, "is_self_signature_allowed must be boolean type"
            end
        end
        local client_certification = cfg["client_certification"]
        if auth_mode == 2 then
            if client_certification == nil or type(client_certification) ~= "table" then
                for k, v in pairs(client_certification) do
                    if pl_path.isfile(v) == false then
                        return false, v.." is not exist"
                    end
                end
            end
        end
    end
    return true, ""
end

local function check_cfg_websocket_device_list(cfg)
    local deviceList = cfg["deviceList"]

    for k, v in pairs(deviceList) do
        pk = v["productKey"]
        dn = v["deviceName"]
        device_alias = v["deviceAlias"]

        utils.check_pk(pk)
        utils.check_dn(dn)
        local code, msg = is_pk_dn_repeated(cfg["deviceList"], pk, dn, k)
        if code == true then
            return false, msg
        end

        if device_alias ~= nil and type(device_alias) ~= "string" then
            return false, "device alias must be string."
        end
    end
    return true
end

local function check_driver_cfg_valid(name, cfg)

    math.randomseed(tonumber(tostring(ngx.now()*1000):reverse():sub(1,6)))
    if cfg == nil then
        return false, "cfg is null"
    end

    if type(cfg["driverConfig"]) ~= "table" or
        type(cfg["serverList"]) ~= "table" or
        type(cfg["deviceList"]) ~= "table" then
        return false, "driverConfig or serverList or deviceList is not in table format"
    end
    if name == "modbus" then
        local code, msg, new_cfg = check_cfg_modbus_server_list(cfg)
        if code == false then
            return false, msg
        end
        code, msg = check_cfg_modbus_device_list(new_cfg)
        if code == false then
            return false, msg
        end

        return true, "ok", new_cfg
    elseif name == "opcua" then
        local code, msg, new_cfg = check_cfg_opcua_server_list(cfg)
        if code == false then
            return false, msg
        end
        code, msg = check_cfg_opcua_device_list(new_cfg)
        if code == false then
            return false, msg
        end

        return true, "ok", new_cfg
    elseif name == "websocket" then
        local code, msg = check_cfg_websocket_server_list(cfg)
        if code == false then
            return false, msg
        end
        code, msg = check_cfg_websocket_device_list(cfg)
        if code == false then
            return false, msg
        end
        return true, "ok", cfg
    end
    return true, "ok", cfg
end

if ("POST" == request_method) then
    local utils_header = require("utils_header")
    local params = utils_header.check_public_boby_data()
    local driver_name = params["DriverName"]
    local config = params["DriverConfig"]
    if (not (driver_name and #driver_name > 0 and config and "table" == type(config))) then
        utils.error_exit(400, "driver name is null or driver config is not json object")
    end

    local ret_code, ret_msg, ret_cfg =  check_driver_cfg_valid(driver_name, config)

    if false == ret_code then
        utils.error_exit(400, "driver config format is invalid: "..ret_msg)
    end
    utils.set_driver_config(driver_name, json.encode(ret_cfg))
    utils.error_exit(200)
end

utils.error_exit(400, "unsupported request method")
