/*
 * Copyright (c) 2018 Alibaba Group Holding Ltd. All rights reserved.
 */

'use strict';

const util = require('util');
const ws = require('./ws_client.js');

const DEFAULT_API_VERSION = 'v1';
const HEADER_AUTH_TOKEN = 'Authorization';
const HEADER_INVOCATION_ID = 'X-Fc-Invocation-Id';
const HEADER_INVOCATION_TYPE = 'X-Fc-Invocation-Type';
const HEADER_INVOKER_CONTEXT = 'X-Fc-Invoker-Context';
const HEADER_FUNCTION_ERROR = 'X-Fc-Function-Error';
const INVOCATION_TYPE_SYNC = 'Sync';
const INVOCATION_TYPE_ASYNC = 'Async';

function getRequestPath(functionId) {
  return util.format('/%s/functions/%s', DEFAULT_API_VERSION, functionId);
}

function getTaskRequestPath(functionId) {
  return util.format('%s/task', getRequestPath(functionId));
}

class IPCWrapper {
  constructor(authToken, host, port) {
    this._authToken = authToken;
  }

  getTask(functionId, callback) {
    var path = getTaskRequestPath(functionId);
    var options = {
      method: 'GET',
      path: path,
      headers: {
        [HEADER_AUTH_TOKEN]: this._authToken,
      },
    };
    console.log(`Getting task for function "${functionId}" from "${path}".`);
    ws.request(options, (err, res) => {
      if (err) {
        err = new Error(`Failed to get task: ${err}`);
        console.error(err);
        callback(err);
        return;
      }
      if (res.statusCode !== 200) {
        err = new Error(`Get task return status code: "${res.statusCode}".`);
        callback(err);
        return;
      }
      var invocationId = res.headers[HEADER_INVOCATION_ID];
      var invokerContext = res.headers[HEADER_INVOKER_CONTEXT];
      console.log(`Got task item with invocation id "${invocationId}".`);
      callback(null, invocationId, invokerContext, res.body);
    });
  }

  postTaskResult(functionId, invocationId, result, callback) {
    var path = getTaskRequestPath(functionId);
    var options = {
      method: 'POST',
      path: path,
      headers: {
        [HEADER_AUTH_TOKEN]: this._authToken,
        [HEADER_INVOCATION_ID]: invocationId,
      },
      body: result,
    };
    console.log(`Posting task result for invocation id "${invocationId}" to "${path}".`);
    ws.request(options, (err, res) => {
      if (err) {
        err = new Error(`Failed to post task result: ${err}`);
        console.error(err);
        callback(err);
        return;
      }
      if (res.statusCode !== 200) {
        err = new Error(`Post task result return status code: "${res.statusCode}".`);
        callback(err);
        return;
      }
      console.log(`Posted task result for invocation id "${invocationId}".`);
      callback(undefined);
    });
  }

  postHandlerError(functionId, invocationId, error, callback) {
    var path = getTaskRequestPath(functionId);
    var options = {
      method: 'POST',
      path: path,
      headers: {
        [HEADER_AUTH_TOKEN]: this._authToken,
        [HEADER_INVOCATION_ID]: invocationId,
        [HEADER_FUNCTION_ERROR]: 'Handled',
      },
      body: error,
    };
    console.log(`Posting task error for invocation id "${invocationId}" to "${path}".`);
    ws.request(options, (err, res) => {
      if (err) {
        err = new Error(`Failed to post task error: ${err}`);
        console.error(err);
        callback(err);
        return;
      }
      if (res.statusCode !== 200) {
        err = new Error(`Post task error return status code: "${res.statusCode}".`);
        callback(err);
        return;
      }
      console.log(`Posted task error for invocation id "${invocationId}".`);
      callback(undefined);
    });
  }

  postTask(functionId, invocationType, invokerContext, payload, callback) {
    if (invocationType !== INVOCATION_TYPE_SYNC && invocationType !== INVOCATION_TYPE_ASYNC) {
      callback(new Error(util.format('InvocationType "%s" is incorrect, should be "%s"' +
        ' or "%s".', invocationType, INVOCATION_TYPE_SYNC, INVOCATION_TYPE_ASYNC)));
      return;
    }
    var path = getRequestPath(functionId);
    var options = {
      method: 'POST',
      path: path,
      headers: {
        [HEADER_AUTH_TOKEN]: this._authToken,
        [HEADER_INVOKER_CONTEXT]: invokerContext,
        [HEADER_INVOCATION_TYPE]: invocationType,
      },
      body: payload,
    };
    console.log(`Posting task for function "${functionId}" to "${path}".`);
    ws.request(options, (err, res) => {
      if (err) {
        err = new Error(`Failed to post task: ${err}`);
        console.error(err);
        callback(err);
        return;
      }
      if (res.statusCode !== 200) {
        console.log(`Get task result return status code ${res.statusCode}.`);
      /*
        err = new Error(`Get task result return status code ${res.statusCode}.`);
        console.error(err);
        callback(err);
        return;
      */
      }
      var functionError = res.headers[HEADER_FUNCTION_ERROR];
      callback(undefined, res.body, functionError, res.statusCode);
      /*
      if ((invocationType === INVOCATION_TYPE_SYNC && res.statusCode !== 200)
        || (invocationType === INVOCATION_TYPE_ASYNC && res.statusCode != 202)) {
        err = new Error(util.format('Post task returned with invocation type "%s",' +
          ' status code "%s".', invocationType, res.statusCode));
        console.error(err);
        callback(err);
        return;
      }
      var invocationId = res.headers[HEADER_INVOCATION_ID];
      console.log(`Task posted with invocation id "${invocationId}".`);
      callback(null, invocationId);
      */
    });
  }

  getTaskResult(functionId, invocationId, callback) {
    var path = getRequestPath(functionId);
    var options = {
      method: 'GET',
      path: path,
      headers: {
        [HEADER_AUTH_TOKEN]: this._authToken,
        [HEADER_INVOCATION_ID]: invocationId,
      },
    };
    console.log(`Getting task result for invocation id "${invocationId}" from "${path}".`);
    ws.request(options, (err, res) => {
      if (err) {
        err = new Error(`Failed to get task result: ${err}`);
        console.error(err);
        callback(err);
        return;
      }
      if (res.statusCode !== 200) {
        err = new Error(`Get task result return status code ${res.statusCode}.`);
        console.error(err);
        callback(err);
        return;
      }
      var functionError = res.headers[HEADER_FUNCTION_ERROR];
      callback(undefined, res.body, functionError, res.statusCode);
    });
  }
}

module.exports = IPCWrapper;
